"""setting module for your project"""
import logging

logging.basicConfig(level=logging.INFO)

# your ant`s class modules or packages
ANT_PACKAGES = ['ants']

# custom setting, eg:
# MYSQL_HOST = '127.0.0.1'
